library tencent_flutter_txcloud;
import 'dart:convert';
import 'dart:io';
import 'package:dio/dio.dart';
import 'dart:typed_data';
import 'package:intl/intl.dart';
import 'package:crypto/crypto.dart';
import 'package:hex/hex.dart';
import 'package:tencent_flutter_txcloud/model/replay_data.dart';

/// RequestVideoData类用于从腾讯云拉取回放视频的数据，包括视频名字，时长，URL地址等信息
/// 通过getDataList()方法可以获取到经过解析的数据（QClassReplayPageData对象)，也可以
/// 通过requestDataByPost()方法获取response原始数据后自行解析
/// RequestVideoData类中的其他方法都是私有的对外不可见，这些方法主要是用于计算签名以及
/// 拼接post请求所需要的字符串
///
/// RequestVideoData的构造方法的参数表示post请求的请求体，
/// 里面可以放一些搜索和返回限制条件，例如根据创建时间升序排序，拉去回放数量等等，
/// 全部参数见 https://cloud.tencent.com/document/api/266/31813
/// 构造函数的参数默认值是"{\"Limit\":100}"，表示拉取腾讯云中最多100条符合搜索条件数据，
/// 这里无搜索条件，相当于所有记录都符合条件
///
/// Example：
///     String data =
///         '{\"Limit\":50,\"CreateTime\":{\"After\":\"2021-05-08T06:01:52Z\",\"Before\":\"2021-05-08T07:25:52Z\"}}';
///     RequestVideoData requestVideoData = new RequestVideoData(
///         'RP7anjX9QbdRmupi39******************',
///         'AKIDg7QUIpWAygPsIS******************',
///         body: data);
///     List<ReplayVideoData> dataList = new List();
///     dataList = await requestVideoData.getDataList();

class RequestVideoData {
  final String secretKey;
  final String secretId;
  String body;

  ///回放的总数量
  int _totalCount = 0;
  /// 构造函数
  ///
  /// @param secretId，secretKey 分别是在云API密钥上申请的标识身份的SecretId，
  /// 一个SecretId 对应唯一的SecretKey，而SecretKey会用来生成请求签名 Signature。
  /// 密钥对为必选参数
  ///
  /// @param body 表示post请求的请求体，里面可以放一些搜索和返回限制条件，例如根据
  /// 创建时间升序排序，返回符合搜索条件的数量限制，媒体文件分类ID集合等，全部参数见
  /// https://cloud.tencent.com/document/api/266/31813
  /// 参数默认值是"{\"Limit\":100}"，表示返回的记录数最多为100条，没有搜索条件
  /// body的格式需要符合JSON格式，可以直接手写JSON格式，也可以通过Map转换
  ///
  /// Example：
  ///     import 'dart:convert' as convert;
  ///     Map mapData = { 'Limit': 5000};
  ///     String data = convert.jsonEncode(mapData);
  RequestVideoData(this.secretId,this.secretKey,
      {this.body = '{\"Limit\":100}'});

  /// 拉取网络数据，并解析成ReplayVideoData类对象,每个视频对应一个ReplayVideoData对象
  /// 如果请求失败如：网络拥塞，密钥不存在或错误等原因将会返回一个空的List，如果是
  /// 腾讯云中没有回放视频或没有符合筛选条件的视频，也会返回一个空的List
  Future<List<ReplayVideoData>> getDataList() async {
    List<ReplayVideoData> _dataList = new List();
    var response = await requestDataByPost();
    ///请求失败
    if (response.statusCode != HttpStatus.ok) {
      return _dataList;
    }
    Map map = (response.data)['Response'];

    ///在返回的respond中有一个TotalCount字段，该字段是指符合搜索条件的记录总数，但是
    ///该字段不一定等于实际返回的response中的媒体文件数量，因为有Limit参数的限制，
    ///也就是 _totalCount = min{Limit, TotalCount};
    _totalCount = map['MediaInfoSet'].length;

    ///仓库里没有回放视频或者没有符合查询条件的视频
    if (_totalCount == 0) {
      return _dataList;
    }

    ///媒体文件信息列表
    var mediaInfoSet = map['MediaInfoSet'];

    ///填充数据到ReplayVideoData对象
    for (int i = 0; i < _totalCount; i++) {
      Map basicInfoMap = mediaInfoSet[i]['BasicInfo'];
      Map metaDataMap = mediaInfoSet[i]['MetaData'];
      _dataList.add(new ReplayVideoData(
          basicInfoMap['Name'],
          basicInfoMap['CreateTime'],
          basicInfoMap['ClassName'],
          basicInfoMap['CoverUrl'],
          basicInfoMap['Type'],
          basicInfoMap['MediaUrl'],
          basicInfoMap['Vid'],
          basicInfoMap['ClassId'],
          metaDataMap['Size'],
          metaDataMap['Duration'],
          description: basicInfoMap['Description'],
          upDate: basicInfoMap['UpdateTime'],
          expireTime: basicInfoMap['ExpireTime'],
          category: basicInfoMap['Category'],
          videoDuration: metaDataMap['VideoDuration'].toDouble(),
          audioDuration: metaDataMap['AudioDuration'].toDouble(),
          width: metaDataMap['Width'],
          height: metaDataMap['Height']));
    }
    return _dataList;
  }


  /// 由于getDataList()方法只解析了一些常用的应答数据，所以如果需要一些未在前面方法中解析
  /// 的数据，请通过调用requestDataByPost()方法，得到返回的response后自行解析
  ///
  /// Example：
  ///   Map mapData = {
  ///     'Limit': 5000,
  ///     'CreateTime': {
  ///       'After': '2021-05-08T06:01:52Z',
  ///       'Before': '2021-05-08T07:25:52Z'
  ///     }
  ///   };
  ///   String data = convert.jsonEncode(mapData);
  ///   RequestVideoData requestVideoData = new RequestVideoData(
  ///       'AKIDg7QUIpWAygPsI*********************',
  ///       'RP7anjX9QbdRmupi3*****************',
  ///       body: data);
  ///   var response =await requestVideoData.requestDataByPost();
  Future<Response> requestDataByPost() async {
    String timestamp = _getTimeStamp();
    String date = _getDate(timestamp);
    String signature = _getPostSignature(body, timestamp, date);
    var dio = Dio(BaseOptions(
      /// 下面post方法中的url会自动覆盖baseUrl
      /// baseUrl: 'https://vod.tencentcloudapi.com',
      method: 'POST',
      ///时间单位是毫秒
      sendTimeout: 5000,
      connectTimeout: 5000,
      receiveTimeout: 10000,
      ///post请求头部参数，包括一些腾讯云API公共参数
      headers: {
        'X-TC-Action': 'SearchMedia',
        'X-TC-RequestClient': 'APIExplorer',
        'X-TC-Timestamp': timestamp,
        'X-TC-Version': '2018-07-17',
        'X-TC-Language': 'zh-CN',
        'Content-Type': 'application/json',
        'Authorization':
        'TC3-HMAC-SHA256 Credential=$secretId/$date/vod/tc3_request, SignedHeaders=content-type;host, Signature=$signature',
      },

      /// 如果使用dart中自带的枚举类型JASON作为contentType的值会出问题，自带的JASON会
      /// 自带 charset=utf-8字段，会导致签名出错，需要手动赋值：application/json
      contentType: 'application/json',
      /// 这这儿设置了返回类型为json，得到数据后就不需要再用jason解码器解码，不然会出错
      responseType: ResponseType.json,
    ));

    /// 发起post请求，会返回一个Response类型的Future对象
    Response response =
    await dio.post('https://vod.tencentcloudapi.com', data: body);
    print(response.data);
    return response;
  }

  /// 对传入的http请求正文做 SHA256哈希，并进行十六进制编码（返回的Digest类型已经是十六进制编码）
  String _sha256Hex(String data) {
    var bytes = utf8.encode(data);
    var digest = sha256.convert(bytes);
    return digest.toString().toLowerCase();
  }

  /// HMAC加密算法，一种非对称加密算法
  List<int> _hmac256(List<int> key, String data) {
    var bytes = utf8.encode(data);
    var hMacSha256 = new Hmac(sha256, key);
    var sha256Result = hMacSha256.convert(bytes);

    /// 将加密的结果转换成字节编码，以免在后面的加密计算中造成加密结果不一致从而导致签名计算错误
    /// 具体原因个人认为（不一定正确）是由于dart中只有int和double类型，且默认是64位造成
    var byteyList = Int8List.fromList(sha256Result.bytes);
    return byteyList;
  }

  /// 获得当前系统的时间戳，用于计算签名与拼接post请求的头部
  String _getTimeStamp() {
    return (DateTime.now().millisecondsSinceEpoch ~/ 1000).toString();
  }

  /// 根据系统时间戳获得utf标准时间，用于计算签名与拼接post请求的头部
  String _getDate(String timestamp) {
    var sdf = new DateFormat('yyyy-MM-dd');

    /// 注意时区，否则容易出错,设置成utc时区
    String date = sdf.format(new DateTime.fromMillisecondsSinceEpoch(
        int.parse(timestamp + '000'),
        isUtc: true));
    return date;
  }

  /// 生成请求签名,这里采用的是腾讯云推荐的v3签名方法，具体的签名计算过程参见
  /// https://cloud.tencent.com/document/api/266/31757
  String _getPostSignature(String body, String timestamp, String date) {
    String service = "vod";
    String host = "vod.tencentcloudapi.com";
    String algorithm = "TC3-HMAC-SHA256";

    /// 1. 拼接规范请求串
    String httpRequestMethod = "POST";
    /// API3.0固定为正斜杠"/"
    String canonicalUri = "/";
    /// 发起HTTP请求的URL中的查询字符串，对于POST请求，固定为空字符串""
    String canonicalQueryString = "";
    /// 此处不能直接参照原文的演示示例，需要删除"；charset=utf-8" 字段，
    /// 不然也会计算计算签名失败，这是经过多次失败后得到的痛苦教训，不过原文中也有说明
    String canonicalHeaders =
        "content-type:application/json\n" + "host:" + host + "\n";
    String signedHeaders = "content-type;host";
    /// 对post请求正文做SHA256哈希
    String hashedRequestPayload = _sha256Hex(body);

    /// 2. 拼接待签名字符串
    String canonicalRequest = httpRequestMethod +
        "\n" +
        canonicalUri +
        "\n" +
        canonicalQueryString +
        "\n" +
        canonicalHeaders +
        "\n" +
        signedHeaders +
        "\n" +
        hashedRequestPayload;
    String credentialScope = date + '/' + service + '/' + "tc3_request";
    String hashedCanonicalRequest = _sha256Hex(canonicalRequest);
    String stringToSign = algorithm +
        '\n' +
        timestamp +
        '\n' +
        credentialScope +
        '\n' +
        hashedCanonicalRequest;

    /// 3. 计算签名
    var secretDate = _hmac256(utf8.encode('TC3' + secretKey), date);
    var secretService = _hmac256(secretDate, service);
    var secretSigning = _hmac256(secretService, 'tc3_request');
    var signature = _hmac256(secretSigning, stringToSign);
    /// 签名计算是需要byte数组在dart中用int8来转换，但是最终签名需要十六进制字符串，
    /// 直接转换会出问题，签名字符串中会出现负号("-")(前面的生成的是byte数组，byte
    /// 的范围是-128～127，需要将其转换成Uint8再转换成字符串，相当于是对于负数求
    /// 其补码后再进行转换。
    var decode = Uint8List.fromList(signature);
    /// 获取一个解码器，将签名数组转换成十六进制字符串
    var encoder = HexEncoder();
    var signatureString = encoder.convert(decode);
    return signatureString;
  }

}
